/*

  Gatekeeper Central Server for A0B36APO Subject
  Czech technical University in Prague
  Copyright 2014, Pavel Pisa <pisa@cmp.felk.cvut.cz>

  Code based on uLAN ETHERNET daemon (ul_ethd)
  by Petr Smolik and Pavel Pisa and other GPL/MPL based
  projects.
*/

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <getopt.h>
#include <sys/types.h>
#ifdef _WIN32
 #include <winsock.h>
 #include <fcntl.h>
 typedef int socklen_t;
#else
 #include <sys/select.h>
 #include <sys/socket.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
 #define closesocket(s)  close(s)
#endif /* _WIN32 */
#include <sys/stat.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <ctype.h>
#include <errno.h>

#define HAS_GETOPT_LONG
#define HAS_GETDELIM

#include "utils.h"
#include "gatecentral.h"

/*******************************************************************/
/* Global variables */

const char *gatec_config_filename = "gatecentral.cfg";
int gatec_server_port = 55556;
int gatec_server_socket;
int gatec_server_terminated;
int verbose_fl;

gatec_config_t *gatec_config;

/*******************************************************************/
/* Utility and portability functions */

#ifndef HAS_GETDELIM
int getdelim(char **line,size_t *linelen,char delim,FILE *F)
{
 char c;
 size_t  l=0;
 do{
  if(l+1>=*linelen)
  {
   *linelen=l+20;
   if(!*line) *line=(char *)malloc(*linelen);
    else *line=(char *)realloc(*line,*linelen);
  }
  c=fgetc(F);
  if(feof(F)) {if(l) break; else return -1;}
  if(c!='\r') (*line)[l++]=c;
 } while(c!=delim);
 (*line)[l]=0;
 return l;
}
#endif

void line_rtrim(char *line)
{
  size_t l = strlen(line);
  while (l) {
    if(!strchr("\n\r \t", line[l - 1]))
      break;
    l--;
    line[l] = 0;
  }
}

/*******************************************************************/
/* Configuration file read */

gatec_user_t *gatec_config_user_find(gatec_config_t *cfg, int id)
{
  gatec_user_t *user = cfg->users;

  while (user) {
    if (user->user_id == id)
      return user;
    user=user->next;
  }
  return NULL;
}

gatec_gate_t *gatec_config_gate_find(gatec_config_t *cfg, int id)
{
  gatec_gate_t *gate = cfg->gates;

  while (gate) {
    if (gate->gate_id == id)
      return gate;
    gate=gate->next;
  }
  return NULL;
}

void gatec_config_destroy(gatec_config_t *cfg)
{
  gatec_user_t *user;
  gatec_gate_t *gate;
  void *p;

  if (cfg == NULL)
    return;

  user = cfg->users;
  gate = cfg->gates;

  while(user) {
    p = user;
    user = user->next;
    free(p);
  }

  while(gate) {
    if (gate->allowed)
      free(gate->allowed);
    p = gate;
    gate = gate->next;
    free(p);
  }

  free(cfg);
}

#define GATEC_CONFIG_TOKEN_MAX 20

gatec_config_t *gatec_config_read(const char *config_filename)
{
  char *p;
  char *line = NULL;
  size_t line_len = 0;
  int line_num = 0;
  FILE *file;
  gatec_config_t *cfg;
  char token[GATEC_CONFIG_TOKEN_MAX+1];
  gatec_user_t *user = NULL;
  gatec_gate_t *gate = NULL;

  cfg = (gatec_config_t *)malloc(sizeof(*cfg));
  if (cfg == NULL) {
    fprintf(stderr, "gatec_config_read: out of memory\n");
    return NULL;
  }
  memset(cfg, 0, sizeof(*cfg));

  file = fopen(config_filename, "r");
  if (!file) {
    fprintf(stderr, "gatec_config_read: file \"%s\" open error - %s\n",
            config_filename, strerror(errno));
    goto error_file;
  }

  while(1) {
      line_num++;
      if(getdelim(&line, &line_len, '\n', file) == -1)
        break;
      if (line == NULL)
        break;
      line_rtrim(line);
      p=line;
      si_skspace(&p);
      if (!*p || (*p == '#'))
        continue;
      if (si_alnumn(&p, token, GATEC_CONFIG_TOKEN_MAX) <= 0) {
        fprintf(stderr, "gatec_config_read: line %d does not start with token \"%s\"\n",
            line_num, line);
        goto error_parse;
      }
      if (!strcmp(token, "gate")) {
        long id;
        si_skspace(&p);
        if (si_long(&p, &id, 0) <= 0) {
          fprintf(stderr, "gatec_config_read: gate ID parse error at line %d\n",
                       line_num);
          goto error_parse;
        }
        if (gatec_config_gate_find(cfg, id)) {
          fprintf(stderr, "gatec_config_read: gate id %ld duplicate at line %d\n",
                           id, line_num);
          goto error_parse;

        }
        gate = (gatec_gate_t *)malloc(sizeof(*gate));
        if (gate == NULL) {
          fprintf(stderr, "gatec_config_read: out of memory\n");
          goto error_parse;
        }
        memset(gate, 0, sizeof(*gate));
        gate->gate_id = id;

        si_skspace(&p);
        while (*p) {
          if (si_alnumn(&p, token, GATEC_CONFIG_TOKEN_MAX) <= 0) {
            fprintf(stderr, "gatec_config_read: gate line %d garbage instead of gate config\n",
                    line_num);
            goto error_parse;
          }
          if (strcmp(token, "allowed")) {
            fprintf(stderr, "gatec_config_read: gate %d line %d \"allowed\" token missing\n",
                    gate->gate_id, line_num);
            goto error_parse;
          }
          si_skspace(&p);
          if (*p) {
            int byte_len = gate->allowed_len * sizeof(*gate->allowed);
            if (si_add_to_arr(&p, (void**)&gate->allowed, &byte_len, 0,
                sizeof(*gate->allowed), NULL) < 0) {
              fprintf(stderr, "gatec_config_read: gate %d line %d allowed users parse error \"%s\"\n",
                    gate->gate_id, line_num, p);
              goto error_parse;
            }
            gate->allowed_len = byte_len / sizeof(*gate->allowed);
          }
          si_skspace(&p);
        }
        gate->next = cfg->gates;
        cfg->gates = gate;
        gate = NULL;
      } else if (!strcmp(token, "user")) {
        long id, pin;
        si_skspace(&p);
        if (si_long(&p, &id, 0) <= 0) {
          fprintf(stderr, "gatec_config_read: user ID parse error at line %d\n",
                       line_num);
          goto error_parse;
        }
        if (gatec_config_user_find(cfg, id)) {
          fprintf(stderr, "gatec_config_read: user id %ld duplicate at line %d\n",
                           id, line_num);
          goto error_parse;

        }
        user = (gatec_user_t *)malloc(sizeof(*user));
        if (user == NULL) {
          fprintf(stderr, "gatec_config_read: out of memory\n");
          goto error_parse;
        }
        memset(user, 0, sizeof(*user));
        user->user_id = id;

        si_skspace(&p);
        while (*p) {
          if (si_alnumn(&p, token, GATEC_CONFIG_TOKEN_MAX) <= 0) {
            fprintf(stderr, "gatec_config_read: user line %d garbage instead of user config\n",
                    line_num);
            goto error_parse;
          }
          if (strcmp(token, "pin")) {
            fprintf(stderr, "gatec_config_read: user %d line %d \"pin\" token missing\n",
                    user->user_id, line_num);
            goto error_parse;
          }
          si_skspace(&p);
          if (si_long(&p, &pin, 0) <= 0) {
            fprintf(stderr, "gatec_config_read: user %d pin parse error at line %d\n",
                            user->user_id, line_num);
            goto error_parse;
          }
          user->pin = pin;
          si_skspace(&p);
        }

        user->next = cfg->users;
        cfg->users = user;
        user = NULL;
      } else {
        fprintf(stderr, "gatec_config_read: unknown token \"%s\" at line %d\n",
            token, line_num);
        goto error_parse;
      }
  }

  fclose(file);

  if (line != NULL)
    free(line);

  return cfg;

error_parse:
  if (user != NULL)
    free(user);
  if (gate != NULL) {
    if (gate->allowed != NULL)
      free(gate->allowed);
    free(gate);
  }

error_file:
  gatec_config_destroy(cfg);

  if (line != NULL)
    free(line);

  return NULL;
}

/*******************************************************************/
/* Clients requests processing */

#define GATEC_CLIENT_TOKEN_MAX 20

void* gatec_client_run(void *vcontext)
{
  gatec_client_context_t *client=(gatec_client_context_t *)vcontext;
  int fd = client->sock;
  char token[GATEC_CLIENT_TOKEN_MAX+1];
  gatec_user_t *user = NULL;
  gatec_gate_t *gate = NULL;
  char *p;
  char *line = NULL;
  size_t line_len = 0;
  int r = -1;

  if (verbose_fl)
    fprintf(stderr, "gatecentral: thread for client fd:%d is running\n",fd);

  while (1) {

    if (client->terminated || feof(client->sock_file)) {
      r=0;
      break;
    }

    fflush(client->sock_file);

    if(getdelim(&line, &line_len, '\n', client->sock_file) == -1) {
      if (feof(client->sock_file))
        continue;
      fprintf(stderr, "gatecentral: client fd:%d getdelim error %s\n",
                      fd, strerror(errno));
      break;
    }
    if (line == NULL) {
      fprintf(stderr, "gatecentral: client fd:%d getdelim empty line\n", fd);
      break;
    }
    line_rtrim(line);
    p=line;
    si_skspace(&p);
    if (!*p || (*p == '#'))
      continue;

    if (verbose_fl)
      fprintf(stderr, "gatecentral: client fd:%d request \"%s\"\n", fd, line);

    if (si_alnumn(&p, token, GATEC_CLIENT_TOKEN_MAX) <= 0) {
      fprintf(stderr, "gatecentral: client fd:%d request missing token \"%s\"\n",
                      fd, line);
      fprintf(client->sock_file, "error hint \"request missing token\"\n");
      break;
    }

    if (!strcmp(token,"checkaccess")) {
      long gate_id;
      long user_id;
      long pin;
      int i;
      gatec_config_t *cfg;

      if (si_long(&p, &gate_id, 0) <= 0) {
        fprintf(stderr, "gatecentral: client fd:%d checkaccess request missing gate ID\n",
                         fd);
        fprintf(client->sock_file, "error hint \"missing gate ID\"\n");
        continue;
      }

      if (si_long(&p, &user_id, 0) <= 0) {
        fprintf(stderr, "gatecentral: client fd:%d checkaccess request missing user ID\n",
                         fd);
        fprintf(client->sock_file, "error hint \"missing user ID\"\n");
        continue;
      }

      if (si_long(&p, &pin, 0) <= 0) {
        fprintf(stderr, "gatecentral: client fd:%d checkaccess request missing user PIN\n",
                         fd);
        fprintf(client->sock_file, "error hint \"missing user PIN\"\n");
        continue;
      }

      cfg = gatec_config;

      user = gatec_config_user_find(cfg, user_id);
      if (user == NULL) {
        fprintf(stderr, "gatecentral: client fd:%d checkaccess user ID %ld unknown\n",
                       fd, user_id);
        fprintf(client->sock_file, "checkaccess %ld %ld forbidden hint \"user id %ld unknown\"\n",
                       gate_id, user_id, user_id);
        continue;
      }

      gate = gatec_config_gate_find(cfg, gate_id);
      if (gate == NULL) {
        fprintf(stderr, "gatecentral: client fd:%d checkaccess gate ID %ld unknown\n",
                       fd, gate_id);
        fprintf(client->sock_file, "checkaccess %ld %ld forbidden hint \"gate id %ld unknown\"\n",
                       gate_id, user_id, gate_id);
        continue;
      }

      if (user->pin != pin) {
        if (verbose_fl)
          fprintf(stderr, "gatecentral: client fd:%d checkaccess user ID %ld PIN missmatch\n",
                       fd, user_id);
        fprintf(client->sock_file, "checkaccess %ld %ld forbidden hint \"user id %ld pin missmatch\"\n",
                       gate_id, user_id, user_id);
        continue;
      }

      i = gate->allowed_len;
      while(i--) {
        if (gate->allowed[i] == user_id)
          break;
      }
      if (i < 0) {
        if (verbose_fl)
          fprintf(stderr, "gatecentral: client fd:%d checkaccess user ID %ld not allowd to gate ID %ld\n",
                       fd, user_id, gate_id);
        fprintf(client->sock_file, "checkaccess %ld %ld forbidden hint \"user id %ld not allowed for gate id %ld\"\n",
                       gate_id, user_id, user_id, gate_id);
        continue;
      }

      fprintf(client->sock_file, "checkaccess %ld %ld OK\n", gate_id, user_id);

    } else {
      fprintf(stderr, "gatecentral: client fd:%d send unknown token \"%s\"\n",
                      fd, token);
      fprintf(client->sock_file, "error hint \"unknown token %s\"\n", token);
      continue;
    }
  }
  if (verbose_fl)
    fprintf(stderr, "gatecentral: thread for client fd:%d terminated by rcode:%d\n",fd,r);

  fflush(client->sock_file);
  fclose(client->sock_file);

  if (r==-1) {
    if (verbose_fl)
      fprintf(stderr, "gatecentral: closing socket in thread\n");

    /* closesocket(client->sock); */
    /* cannot be called, socket alread closed by fclose, but windows are question */

    /* terminated by client side request */
    /*
    pthread_mutex_lock (&client_mutex);
    client_delete(&client_root,client);
    pthread_mutex_unlock (&client_mutex);
    */
  }
  free(client);

  if (line)
    free(line);

  if (verbose_fl)
    fprintf(stderr, "gatecentral: thread for client fd:%d was terminated\n",fd);

  pthread_detach(pthread_self());
  pthread_exit(NULL);
  return NULL;
}


/*******************************************************************/
/* Remote requests server */

int gatec_server_init(void)
{
  struct sockaddr_in address;
  int server_socket;
  int yes=1;
 #ifdef _WIN32
  WORD wVersionRequested;
  WSADATA wsaData;
  int sockopt = SO_SYNCHRONOUS_NONALERT;
 #endif

  gatec_server_terminated = 0;

 #ifdef _WIN32
  wVersionRequested = MAKEWORD(2, 0);
  WSAStartup(wVersionRequested, &wsaData);
  setsockopt(INVALID_SOCKET, SOL_SOCKET, SO_OPENTYPE, (char *)&sockopt, sizeof(sockopt));
 #endif

  if ((server_socket = socket(AF_INET,SOCK_STREAM,0)) < 0) {
    fprintf(stderr, "gatecental: the socket created error\n");
    exit(1);
  }
  if (verbose_fl)
    fprintf(stderr, "gatecental: the server socket fd:%d was created\n", server_socket);

  /* Make listening socket's port reusable - must appear before bind */
  if (setsockopt(server_socket, SOL_SOCKET, SO_REUSEADDR, (char*)&yes,sizeof(yes)) < 0) {
    fprintf(stderr, "gatecental: setsockopt failure\n");
    exit(1);
  }

  address.sin_family = AF_INET;
  address.sin_addr.s_addr = INADDR_ANY;
  address.sin_port = htons(gatec_server_port);
  if (bind(server_socket,(struct sockaddr *)&address,sizeof(address)) < 0) {
    fprintf(stderr, "gatecental: bind error\n");
    exit(1);
  }
  if (verbose_fl)
    fprintf(stderr, "gatecental: server socket fd:%d bound to port %d \n", server_socket, gatec_server_port);

  /*  Create a connection queue and initialize readfds to handle input from server_socket. */
  if(listen(server_socket, 5) < 0) {
    fprintf(stderr, "gatecental: error to listen on socket\n");
    exit(1);
  }

  gatec_server_socket = server_socket;

  return 0;
}

int gatec_server_run(void)
{
  int csock;
  fd_set readfds, testfds;
  int result, f;
  socklen_t addrlen;
  struct sockaddr_in address;
  int server_socket = gatec_server_socket;

  if (verbose_fl)
    fprintf(stderr, "gatecental: server run\n");

  FD_ZERO(&readfds);
  FD_SET(server_socket, &readfds);

  while(!gatec_server_terminated) {
    struct timeval timeout = {.tv_sec = 5, .tv_usec = 0};

    testfds=readfds;
    result = select(FD_SETSIZE, &testfds, (fd_set *)0, (fd_set *)0, &timeout);

    if(result < 0) {
      break;
    }

    if (result == 0)
      continue;

    if(FD_ISSET(server_socket, &testfds)) { /*  request for a new connection */
      addrlen = sizeof(struct sockaddr_in);
      csock = accept(server_socket, (struct sockaddr *)&address, &addrlen);
      if (csock > 0){
        gatec_client_context_t *client;
        if (verbose_fl)
          fprintf(stderr, "gatecental: the client from %s is connected, fd:%d\n",
                          inet_ntoa(address.sin_addr), csock);

        client = (gatec_client_context_t *)malloc(sizeof(gatec_client_context_t));
        if (client == NULL) {
          fprintf(stderr, "gatecental: could not get memory\n");
          exit(1);
        }
        memset(client, sizeof(*client), 0);
        client->sock = csock;
        client->terminated = 0;

       #ifndef _WIN32
        f = client->sock;
       #else
        f = _open_osfhandle(client->sock, _O_WRONLY);
        if (f < 0) {
          fprintf(stderr, "gatecental: error _open_osfhandle\n");
          exit(1);
        }
       #endif /* _WIN32 */

        client->sock_file = fdopen(f, "r+");
        if (!client->sock_file) {
          fprintf(stderr, "gatecental: error fdopen fd:%d\n", client->sock);
          exit(1);
        }

        /*
        pthread_mutex_lock (&client_mutex);
        client_insert(&client_root, client);
        pthread_mutex_unlock (&client_mutex);
        */

        if(pthread_create(&client->threadid, NULL, gatec_client_run, (void*)client) != 0){
          fprintf(stderr, "gatecental: could not create client thread\n");
          exit(1);
        }
      }
    } else {
      // this should never happen if the program work well
      fprintf(stderr, "gatecental: ERROR: read from unregistered fd\n");
      exit(1);
    }
  }

  return 0;
}

void gatec_server_stop(void)
{
  gatec_server_terminated = 1;
}

/*******************************************************************/
/* Program main/startup/clenup and command line parsing */

void
sigexit(int sig)
{
  fprintf(stderr, "gatecental: signal %d received\n", sig);
  gatec_server_stop();
}

static void
usage(void)
{
  printf("Usage: gatecentral <parameters>\n");
  printf("  -c, --config <filename>  n device [%s]\n", gatec_config_filename);
  printf("  -p, --port <value>       server port [%d]\n", gatec_server_port);
  printf("  -v, --verbose            verbose output\n");
  printf("  -V, --version            show version\n");
  printf("  -h, --help               this usage screen\n");
}

int main(int argc,char *argv[])
{
  static struct option long_opts[] = {
    { "config", 1, 0, 'c' },
    { "port", 1, 0, 'p' },
    { "verbose", 0, 0, 'v' },
    { "version",0,0, 'V' },
    { "help",  0, 0, 'h' },
    { 0, 0, 0, 0}
  };
  int opt;

 #ifndef HAS_GETOPT_LONG
  while ((opt = getopt(argc, argv, "c:p:vVh")) != EOF)
 #else
  while ((opt = getopt_long(argc, argv, "c:p:vVh",
                            &long_opts[0], NULL)) != EOF)
 #endif
    switch (opt) {
    case 'c':
      gatec_config_filename = optarg;
      break;
    case 'p':
      gatec_server_port = strtol(optarg,NULL,0);
      break;
    case 'v':
      verbose_fl = 1;
      break;
    case 'V':
      fputs("gatecentral 0.1 (APO 2014)\n", stdout);
      exit(0);
    case 'h':
    default:
      usage();
      exit(opt == 'h' ? 0 : 1);
    }

  gatec_config = gatec_config_read(gatec_config_filename);
  if (gatec_config == NULL) {
    fprintf(stderr, "gatecentral: gonfig file \"%s\" not parsed\n",
            gatec_config_filename);
    exit(1);
  }

  gatec_server_init();
  signal(SIGINT,  sigexit);
  signal(SIGTERM,  sigexit);
  gatec_server_run();
  if (verbose_fl)
    fprintf(stderr, "gatecental: program finishing\n");
  gatec_config_destroy(gatec_config);
  return 0;
}
